# liffile/__init__.py
# pylint: skip-file

from .liffile import *
from .liffile import __all__, __doc__, __version__

# constants are repeated for documentation

__version__ = __version__
"""Liffile version string."""
